from torchok.models.poolings.classification.pooling import (
    Pooling,
)
from torchok.models.poolings.classification.linear import (
    PoolingLinear,
)


__all__ = [
    'Pooling',
    'PoolingLinear'
]
