from torchok.data.datasets.examples.cifar10 import CIFAR10
from torchok.data.datasets.examples.sop import SOP
from torchok.data.datasets.examples.sweet_pepper import SweetPepper


__all__ = [
    'CIFAR10',
    'SOP',
    'SweetPepper',
]
